/* ----------------------------------------------------------------------------
 *     Copyright (c) 2000-2019, Nexus Electronics Ltd (www.nexus.co.uk)
 *                            All rights reserved
 *
 * Redistribution and use in source and derived forms, with or without
 * modification, are permitted provided that the following conditions are met:
 *
 *  * Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 *  * Redistributions in derived forms must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 *  * Neither the name of Nexus Electronics Ltd. nor the names of its
 *    contributors may be used to endorse or promote products derived from
 *    this software without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY NEXUS ELECTRONICS LTD. "AS IS" AND ANY EXPRESS
 * OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
 * WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
 * DISCLAIMED. IN NO EVENT SHALL NEXUS ELECTRONICS LTD. BE LIABLE FOR ANY
 * DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
 * (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
 * SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
 * CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH
 * DAMAGE.
 * ----------------------------------------------------------------------------
 * Title       : triple-DES, key calculation
 * Project     : MAS
 * RCS ID      : 
 * $Id: key.v,v 1.2 2001/01/16 18:13:52 el Exp $
 * ----------------------------------------------------------------------------
 * File        : key.v
 * Author      : E.M. Lavelle
 * Company     : Nexus Electronics Ltd.
 * Last saved  : 22/08/2000
 * Version     : $Revision: 867 $
 * ----------------------------------------------------------------------------
 * Description : 
 * 
 * this module carries out the key transformation, as described on pp272/3 of
 * 'applied cryptography'. it carries out the key permutation, the shifting
 * (note that the shifter only shifts one bit at a time, to minimise the
 * hardware), and the compression permutation. the output is PRTL_KEY.
 * 
 * ----------------------------------------------------------------------------
 * Modification history:
 * Date        Version Author     Comment
 * 22/08/2000  0.1      EML       created
 *
 * ------------------------------------------------------------------------- */

`include "3des_inc.v"           // register decodes

module key(CLK,                 // clock
	   SRST_H,              // sync reset
	   DAV_H,		// register load strobe
	   ADDR,		// register address
	   DATAIN,		// block's input bus
	   ENC_H,               // encode/decode
	   KSEL,		// input key select
	   KLD_H,               // key load strobe, from controller
	   PRTL_KEY             // current 48-bit key output
	   );
   input CLK, SRST_H, DAV_H, ENC_H, KLD_H;
   input  [1:0]  KSEL;
   input  [2:0]  ADDR; 
   input  [1:64] DATAIN;
   output [1:48] PRTL_KEY;

   reg  [1:28] 	 SH_LEFT, SH_RIGHT;
   wire [1:56] 	 SH_KEY, KPERM;
   wire [1:64] 	 KEYIN;
   wire [3:0] 	 KEY_ADDR;
   wire 	 KEYWE_H;
   
   // instantiate the key memory. this is a 64*16 memory, but only the
   // bottom 3 locations are used. the memory is written when DAV_H is
   // active, and ADDR addresses one of the keys. during operation, the
   // required (read) key is selected by KSEL.
   assign KEYWE_H = DAV_H & 
	  ((ADDR == `REG_KEY1) | (ADDR == `REG_KEY2) | (ADDR == `REG_KEY3));
   assign KEY_ADDR = (KEYWE_H)? {1'b0, ADDR}:{2'b0, KSEL};
   keyram key_memory (
	.A(KEY_ADDR),
	.CLK(CLK),
	.D(DATAIN),
	.WE(KEYWE_H),
	.SPO(KEYIN));

   // key permutation: select 56 bits from 64
   assign KPERM =
	  {KEYIN[57], KEYIN[49], KEYIN[41], KEYIN[33], KEYIN[25], KEYIN[17], 
	   KEYIN[9],  KEYIN[1],  KEYIN[58], KEYIN[50], KEYIN[42], KEYIN[34],  
	   KEYIN[26], KEYIN[18], KEYIN[10], KEYIN[2],  KEYIN[59], KEYIN[51],
	   KEYIN[43], KEYIN[35], KEYIN[27], KEYIN[19], KEYIN[11], KEYIN[3],
	   KEYIN[60], KEYIN[52], KEYIN[44], KEYIN[36], KEYIN[63], KEYIN[55],
	   KEYIN[47], KEYIN[39], KEYIN[31], KEYIN[23], KEYIN[15], KEYIN[7],
	   KEYIN[62], KEYIN[54], KEYIN[46], KEYIN[38], KEYIN[30], KEYIN[22],
	   KEYIN[14], KEYIN[6],  KEYIN[61], KEYIN[53], KEYIN[45], KEYIN[37],
           KEYIN[29], KEYIN[21], KEYIN[13], KEYIN[5],  KEYIN[28], KEYIN[20],
           KEYIN[12], KEYIN[4]};

   /************************************************************************** 
    *  left and right shift registers:
    *  SRST_H
    *   KLD_H
    *    ENC_H
    *  1xx           sync reset to 0
    *  01x           load in the new key
    *  000           decode; 1-bit right shift
    *  001           encode; 1-bit left shift
    **************************************************************************/
   always @(posedge CLK) begin : left_reg
      if(SRST_H)
	SH_LEFT <= 0;
      else if(KLD_H)
	SH_LEFT <= KPERM[ 1:28];
      else if(ENC_H)
	SH_LEFT <= {SH_LEFT[2:28], SH_LEFT[1]};
      else
	SH_LEFT <= {SH_LEFT[28], SH_LEFT[1:27]};
   end
   
   always @(posedge CLK) begin : right_reg
      if(SRST_H)
	SH_RIGHT <= 0;
      else if(KLD_H)
	SH_RIGHT <= KPERM[29:56];
      else if(ENC_H)
	SH_RIGHT <= {SH_RIGHT[2:28], SH_RIGHT[1]};
      else
	SH_RIGHT  <= {SH_RIGHT[28], SH_RIGHT[1:27]};
   end

   // compression permutation: this generates the 48-bit output from the
   // 56-bit shifted key
   assign SH_KEY = {SH_LEFT, SH_RIGHT}; 
   assign PRTL_KEY = {SH_KEY[14], SH_KEY[17], SH_KEY[11], SH_KEY[24],
		      SH_KEY[1],  SH_KEY[5],  SH_KEY[3],  SH_KEY[28],
		      SH_KEY[15], SH_KEY[6],  SH_KEY[21], SH_KEY[10],
		      SH_KEY[23], SH_KEY[19], SH_KEY[12], SH_KEY[4],
		      SH_KEY[26], SH_KEY[8],  SH_KEY[16], SH_KEY[7],
		      SH_KEY[27], SH_KEY[20], SH_KEY[13], SH_KEY[2],
		      SH_KEY[41], SH_KEY[52], SH_KEY[31], SH_KEY[37],
		      SH_KEY[47], SH_KEY[55], SH_KEY[30], SH_KEY[40],
		      SH_KEY[51], SH_KEY[45], SH_KEY[33], SH_KEY[48],
		      SH_KEY[44], SH_KEY[49], SH_KEY[39], SH_KEY[56],
		      SH_KEY[34], SH_KEY[53], SH_KEY[46], SH_KEY[42],
		      SH_KEY[50], SH_KEY[36], SH_KEY[29], SH_KEY[32]};

endmodule

/* ----------------------------------------------------------------------------
 * key memory: this is a single-port, 64*3 memory, generated by coregen.
 * ------------------------------------------------------------------------- */

// synthesis translate_off
`include "XilinxCoreLib/C_DIST_MEM_V3_0.v"
// synthesis translate_on

module keyram (A, CLK, D, WE, SPO);
   input [3:0]   A;
   input         CLK;
   input [63:0]  D;
   input 	 WE;
   output [63:0] SPO;

   // synopsys translate_off
   C_DIST_MEM_V3_0 #(
		     4,"0",1,16,1,1,1,1,0,0,0,0,0,0,0,0,0,0,0,1,0,1,
		     "../coregen/keyram.mif",1,1,0,0,0,0,0,0,0,64)
   xilinx_spram_64x16 (.A(A), .CLK(CLK), .D(D), .WE(WE), .SPO(SPO));
   // synopsys translate_on
endmodule

/*----------------------------------- EOF -----------------------------------*/
